
!dmabuffer	= $7FC000	;$3600 bytes used or something around that
!rambase	= $7FFFE0	;where to start storing variables

!currentframe	= !rambase	;2 bytes, current frame #
!mod4counter	= !rambase+2	;1 byte, for step within current frame FF,00,01,02
!frametoggle	= !rambase+3	;1 byte, for current buffer to upload tiles to
!irqflag	= !rambase+4	;1 byte, NMI passed yet?
!throttlestore	= !rambase+5	;1 byte, fraction for framerate throttle
!idleframes	= !rambase+6	;1 byte, frames to wait before proceeding
!brightness	= !rambase+7	;1 byte, low 4 bits for $2100
!irqmirror	= !rambase+8	;1 byte, $4200 settings for IRQ
!irqstep	= !rambase+9	;1 byte, 2 steps per frame
!dma_tableindex	= !rambase+10	;1 byte, current position in DMA table
!bufferaddress	= !rambase+11	;2 bytes, 16bit offset into RAM bank where tiles are about to be read/written
!audiodelay	= !rambase+13	;1 byte, how many frames to delay audio
!videodelay	= !rambase+14	;1 byte, "" video
!fadetimer	= !rambase+15	;1 byte, brightness / fade timer combination
!dma_table	= !rambase+16	;start of table of parameters for each DMA transfer

!testaddress	= !rambase+14	;2 bytes, for whatever

;--

!DTVRAM		= 0
!DTSIZE		= 2
!DTENTRYSIZE	= 4

;--

!AUDIODELAY	= 0
!VIDEODELAY	= 6
!ENDFRAME	= 2325

;-----

SetupVideoAll:
	JSR SetupVideoVariables
	JSR SetupScreen
	JSR SetupVRAM
	JSR SetupWindow
	JSR FX_Init

	RTS

;-----

SetupVideoVariables:
	PHP
	REP #$20
	SEP #$10
	LDX #$00

	STZ.w !currentframe
	;STX.w !mod3counter
	STX.w !frametoggle
	STX.w !irqflag
	STX.w !irqstep
	STX.w !idleframes
	STX.w !fadetimer
	
	LDX #$FF		;mod4 counter must start at FF
	STX.w !mod4counter
	LDX #$80		;screen off for now
	STX.w !brightness
	LDX #$30		;H/V both required for IRQ
	STX.w !irqmirror
	LDX #!AUDIODELAY
	STX.w !audiodelay
	LDX #!VIDEODELAY
	STX.w !videodelay

	PLP
	RTS

;-----

SetupScreen:
	PHP
	REP #$20
	PHD
	LDA #$2100
	TCD
	SEP #$20

	LDA #$80	;because VRAM writes coming up
	STA.b $2100
	LDA #$00	;no HDMA for now
	STA.l $420C
-	LDA.l $4210	;wait until next vblank before actually doing anything
	BPL -

	LDA #$03	;mode 3 display
	STA.b $2105
	LDA #$01	;L1 on main screen
	STA.b $212C
	STZ.b $2126	;no mosaic
	LDA.b #$78+2	;tilemap goes to last 2KB of VRAM, two arranged vertically
	STA.b $2107
	STZ.b $210B

	PLD
	PLP
	RTS

;-----

SetupVRAM:
	PHP
	PHB
	SEP #$20
	REP #$10
	
;intially clear everything out
	LDA #$00	;bank 00 for hardware register
	PHA
	PLB

	LDA #$80	;word writes for initial clear
	STA $2115
	LDX #$0000
	STX $2116

	LDA #$09	;alternate
	STA $4300
	LDA #$18	;$2118
	STA $4301
	LDX.w #.zeroByte
	STX $4302
	LDA.b #.zeroByte>>16 
	STA $4304
	LDX #$0000	;64KB clear
	STX $4305
	LDA #$01
	STA $420B

	PHK		;current bank for that table of values down below
	PLB

;writing the 2.5KB tilemap for a single screen with slightly different arrangements
	LDY #$7B00	;last 2.5KB of VRAM holds the tilemap
	STY $00		;$00 holds current vram row address
	LDX #$0002	;index for the tilemap arrangement table and load the initial entry since it must be preloaded
	REP #$20
	LDA.w .tileArrangementTable
.loopOuter
	LDY.w #240/8	;must move down a row every time the width of tilemap is reached
	PHA		;access next row in tilemap
	LDA $00
	STA.l $2116
	CLC
	ADC #$0020
	STA $00
	PLA
.loop
	STA.l $2118	;store and must check if the tile # has to change
	CMP.w .tileArrangementTable,x
	BNE +
	INX
	INX
	LDA.w .tileArrangementTable,x	;there was a match, load new tile to store
	BMI .break
	INX
	INX
	BRA ++
+	INC A		;next tilemap word AFTER the current has been written, want it to be an inclusive check
++	DEY
	BNE .loop
	BRA .loopOuter	;no need to test loop count, it automatically breaks upon reading the $FFFF
	
.break
	PLB
	PLP
	RTS

;word 0: start tile #
;word 1: end tile #
;FFFF terminates

.tileArrangementTable
	dw $0010,$0257	;buffer 1
	dw $0000,$000F
	dw $0258,$03D7	;buffer 2
	dw $0190,$0257
	dw $0000,$000F
	dw $FFFF

.zeroByte
	db 0

;-----

!WINDOW_LEFTBOUND	= 8
!WINDOW_RIGHTBOUND	= 247

SetupWindow:
	PHP
	SEP #$20

	LDA #$20	;enable window 1 on color
	STA.l $2125
	LDA #$20	;prevent math inside window i.e. dont sub to black in the viewing area
	STA.l $2130
	LDA #$BF	;affect everything, subtract colors
	STA.l $2131
	LDA #$FF	;subtract white i.e. make everything black
	STA.l $2132

	LDA #!WINDOW_LEFTBOUND
	STA.l $2126
	LDA #!WINDOW_RIGHTBOUND
	STA.l $2127

	PLP
	RTS

;-----

ShowVideo:
	PHP
	SEP #$20
	PHB
	LDA.b #!rambase>>16
	PHA
	PLB

	JSR FX_SetupAudio	;this returns only when audio is actually ready to start

-	LDA.l $4210
	BPL -

	JSR UpdateIRQ	;set IRQ v/h to required initial state

	LDA.l $4211
	LDA.w !irqmirror	;ready to service IRQs from now
	ORA #$01		;allow joypad reading too
	STA.l $4200
	CLI
	LDA #$0F	;max brightness from here on
	STA.w !brightness
.loop
	LDA.w !fadetimer	;don't check for exit condition if already fading out
	BNE +
	JSR CheckExitCondition
	BCC +

	LDA #$8F		;bit7=fading out, low nybble = brightness
	STA.w !fadetimer

+	LDA.w !audiodelay	;only play audio when the count is up
	BEQ +
	DEC.w !audiodelay
	BRA ++
+	JSR FX_PlayAudio

++	STZ.w !irqflag
-	LDA.w !irqflag
	BEQ -

	BIT.w !fadetimer	;does fade timer need updating?
	BPL .loop
	DEC.w !fadetimer	;brightness-- and also, testing for exit when the brightness drops to zero
	BPL .return
	LDA.w !fadetimer
	AND #$0F
	STA.w !brightness	;both video/audio go out smoothly
	JSR FX_FadeAudio
	BRA .loop

.return
	JSR FX_StopAudio

	PLB
	PLP
	RTS

;-----

;Out:
;C: = 1 if playback is to stop, 0 otherwise

CheckExitCondition:
	CLC
	PHP
	REP #$20

	LDA.w !currentframe	;do not check if it's the first couple frames, for button presses
	CMP.w #20
	BCC .return
	LDA.l $4218		;any button press at all
	BNE .returnExit
	LDA.w !currentframe	;if currentframe is equal to the fadeout frame
	CMP.w #!ENDFRAME
	BNE .return

.returnExit
	SEP #$20
	LDA $01,s	;carry flag set on terminating condition
	ORA #$01
	STA $01,s	
.return
	PLP
	RTS

;-----

ReadTiles:
	PHP
	REP #$10
	SEP #$20

	LDA.w !mod4counter	;do nothing if this is the first step
	BMI .return

	STZ.w !dma_tableindex	;reset index for DMA table and address to start sending tiles to
	LDX.w #!dmabuffer
	STX.w !bufferaddress

	LDA.w !frametoggle	;this decides what buffer to send stuff to
	ASL A
	CLC
	ADC.w !frametoggle	;x3
	CLC
	ADC.w !mod4counter	;figure out target address this frame
	REP #$20
	AND #$00FF
	STA $00		;x6, 3 sets of words per entry
	ASL A
	CLC
	ADC $00
	ASL A
	TAX
;storing of each step
	LDA.l .vramSize1,x
	TAY
	LDA.l .vramDestination1,x
	JSR SetupDMATransfer
	LDA.l .vramSize1+2,x
	TAY
	LDA.l .vramDestination1+2,x
	JSR SetupDMATransfer
	LDA.l .vramSize1+4,x	;third entry might have zero size indicating 'don't transfer'
	BEQ .return
	TAY
	LDA.l .vramDestination1+4,x
	JSR SetupDMATransfer
.return
	PLP
	RTS

;amount to upload depends on frame and destination depends on frame toggle

.vramDestination1
	dw $0200	;buffer 1 step 1
	dw $0E80	;step 2
	dw $FFFF	;step 3

;.vramDestination2
	dw $1B00	;buffer 1 step 4
	dw $2780	;step 5
	dw $FFFF	;step 6

;.vramDestination3
	dw $3200	;buffer 1 step 7
	dw $3E80	;step 8
	dw $0000	;step 9

;--

;.vramDestinationB1
	dw $4B00	;buffer 2 step 1
	dw $5780	;step 2
	dw $FFFF	;step 3

;.vramDestinationB2
	dw $6400	;buffer 2 step 4
	dw $7080	;step 5
	dw $FFFF	;step 6

;.vramDestinationB3
	dw $3200	;buffer 2 step 7
	dw $3E80	;step 8
	dw $0000	;step 9

;---

.vramSize1
	dw $1900
	dw $1900
	dw 0

;.vramSize2
	dw $1900
	dw $1500
	dw 0

;.vramSize3
	dw $1900
	dw $1900
	dw $0400
	
;--

.vramSizeB1
	dw $1900
	dw $1900
	dw 0

;.vramSizeB2
	dw $1900
	dw $1500
	dw 0

;.vramSizeB3
	dw $1900
	dw $1900
	dw $0400

;-----

;16bit AXY
;A: VRAM destination
;Y: size of transfer

SetupDMATransfer:
	PHP
	PHX
	PHY
	SEP #$10

	LDX.w !dma_tableindex		;VRAM/requested size is all that is required here
	STA.w !dma_table+!DTVRAM,x
	LDA $01,s
	STA.w !dma_table+!DTSIZE,x

	LDA #$FFFF	;terminate table
	STA.w !dma_table+!DTENTRYSIZE,x
	SEP #$20	;adjust index, only a single table was used here
	TXA
	CLC
	ADC.b #!DTENTRYSIZE
	STA !dma_tableindex	;for future runs of this routine

	;....

	REP #$30		;actually get the tiles to transfer into RAM
	LDA $01,s		;grab required size again
	TAY
	LDA.w !bufferaddress	;target RAM address
	LDX.w #!dmabuffer>>16	;bank byte, 16bit load but only low 8 bits get used
	JSR FX_BufferData
	PLA			;size + current buffer address = next buffer address
	CLC
	ADC.w !bufferaddress
	STA.w !bufferaddress

	PLX
	PLP	
	RTS

;-----

VideoIRQ:
	SEP #$30
	LDA.l $4211
	LDA.b #!rambase>>16
	PHA
	PLB

	LDA.w !irqstep		;one of two routines based on current step
	ASL A
	TAX
	INC.w !irqstep
	LDA.w !irqstep
	CMP #$02		;only 00 and 01 are valid here
	BNE +
	STZ.w !irqstep	
+	JSR (.irqJumpTable,x)
	JSR UpdateIRQ

	RTS

.irqJumpTable
	dw IRQ1
	dw IRQ2

IRQ1:
	LDA.w !brightness
	STA.l $2100

	LDA.w !videodelay
	BNE .return

	JSR ReadTiles	;the buffering for graphics data starts here
	JSR ReadPalette
.return
	RTS

IRQ2:
	LDA #$80		;screen disabled in second step..
	STA.l $2100

	LDA.w !videodelay	;don't upload or update counters if video hasn't started yet
	BEQ +
	DEC.w !videodelay
	DEC.w !irqflag
	RTS

+	LDA.w !mod4counter	;don't upload if it's the first step
	BMI +

	JSR UploadDMA

+	DEC.w !irqflag
	INC.w !mod4counter
	LDA.w !mod4counter
	CMP #$03
	BCC .skip
	;STZ.w !mod4counter	;3 frames passed, update whatever necessary
	LDA #$FF
	STA.w !mod4counter
	INC.w !frametoggle
	LDA #$FE
	TRB.w !frametoggle	;only the low bit is of concern
	REP #$20
	INC.w !currentframe
	SEP #$20
	JSR UpdateScroll
.skip
	RTS

;-----

UploadDMA:
	PHP
	PHD
	REP #$20
	SEP #$10
	LDA #$4300
	TCD

;tile portion
	SEP #$20		;regular word writes
	LDA #$80
	STA.l $2115
	REP #$20

	LDA.w #!dmabuffer	;set buffer address to start
	STA.w !bufferaddress
	LDY #$01		;2 regs alternate
	STY.b $4300
	LDY #$18		;$2118
	STY.b $4301
	LDY.b #!dmabuffer>>16
	STY.b $4304		;bank byte remains fixed
	LDX #$00		;read from start of DMA table..
.loop
	LDA.w !dma_table+!DTVRAM,x	;vram / size is a straight copy
	BMI .break
	STA.l $2116
	LDA.w !dma_table+!DTSIZE,x
	STA.b $4305
	LDA.w !bufferaddress		;source address must be updated between transfers
	STA.b $4302
	CLC
	ADC.w !dma_table+!DTSIZE,x
	STA.w !bufferaddress
	SEP #$20
	LDA.b #$01	;i can actually just use the same DP register value for this but I don't want to be THAT tacky
	STA.l $420B
	
	TXA		;access next entry in table
	CLC
	ADC #!DTENTRYSIZE
	TAX
	REP #$20
	BRA .loop

.break

;palette portion (only if mod counter == 2)
	LDX.w !mod4counter
	CPX #$02
	BNE .return

	SEP #$20
	REP #$10
	LDA #$00
	STA.l $2121
	STZ.b $4300		;1 reg only
	LDA #$22		;$2122
	STA.b $4301
	LDY.w !bufferaddress	;source is after all those tiles
	STY.b $4302
	LDY #$0200		;512 byte whole palette
	STY.b $4305
	LDA #$01
	STA.l $420B

.return
	PLD
	PLP
	RTS

;-----

ReadPalette:
	PHP
	SEP #$20

	LDA.w !mod4counter	;final step in this frame? update necessary then
	CMP #$02
	BNE .return

	REP #$30
	LDA.w !bufferaddress	;target RAM address
	LDX.w #!dmabuffer>>16	;bank byte, 16bit load but only low 8 bits get used
	LDY #$0200		;512 byte whole palette
	JSR FX_BufferData	;no adjustment after reading is necessary since this is always the last in any frame

.return
	PLP
	RTS

;-----

!X_SCROLL = -8

UpdateScroll:
	PHP
	SEP #$30
	
;simple load store for scroll:
	LDA.b #!X_SCROLL
	STA.l $210D
	LDA.b #!X_SCROLL>>8
	STA.l $210D

	LDA !frametoggle
	ASL A
	TAX
	LDA.l .yScrollTable,x
	STA.l $210E
	LDA.l .yScrollTable+1,x
	STA.l $210E

	PLP
	RTS

.yScrollTable
	dw 192+160-32-1		;buffer 1
	dw 192-32-1		;buffer 2

;-----

!IRQ_HPOS = 100

UpdateIRQ:
	PHP
	REP #$20
	SEP #$10

	LDA.w #!IRQ_HPOS
	STA.l $4207
	LDA.w !irqstep
	AND #$00FF
	ASL A
	TAX
	LDA.l .irqVPosTable,x
	STA.l $4209

	PLP
	RTS

.irqVPosTable
	dw 32,192	;bottom and top borders

;-----

TestIRQ:
	PHP
	SEP #$20

	LDA #$00
	STA.l $2121
	LDA.w !currentframe
	STA.l $2122
	LDA.w !currentframe+1
	STA.l $2122

	PLP
	RTS